from abc import ABC, abstractmethod

class ValueExtractorBase(ABC):

    @staticmethod
    @abstractmethod
    def can_extract(dicom_obj):
        pass

    @staticmethod
    @abstractmethod
    def extract_value(dicom_obj):
        pass



class ConceptName:
    def __init__(self,
                 code_value: str,
                 code_meaning: str,
                 coding_scheme_designator: str) -> None:

        self.code_value = code_value
        self.code_meaning = snake_case(code_meaning)
        self.coding_scheme_designator = coding_scheme_designator

    def __str__(self) -> str:
        return 'code_value = {0}, coding_scheme_designator = {1}, code_meaning = {2}'.format(
                self.code_value,
                self.coding_scheme_designator,
                self.code_meaning)

    def __repr__(self) -> str:
        return '{0}, {1}, {2}'.format(self.code_value, self.coding_scheme_designator, self.code_meaning)

    @property
    def code_value_csd(self) -> str:
        return '{0}:{1}'.format(self.code_value, self.coding_scheme_designator)

def snake_case(s: str) -> str:
    return s.lower().strip().replace(' ', '_')



from typing import List, Iterator, Union
from rdsr_navigator.concept_name import ConceptName
from rdsr_navigator.repr_display import reper_html
from rdsr_navigator.sr_exceptions import SrMissingContentException
import rdsr_navigator.value_extractors as ves

_VALUE_EXTRACTORS: List[ves.ValueExtractorBase] = [ves.CodeMeaningExtractor(),
                                                   ves.MeasuredValuesExtractor(),
                                                   ves.TextValueExtractor(),
                                                   ves.UidExtractor(),
                                                   ves.DateTimeExtractor()]

class SrElement:
    def __init__(self, dicom_data) -> None:
        self._dicom_data = dicom_data

    def __str__(self) -> str:
        return str(self.concept_name)

    def __repr__(self) -> str:
        return repr(self.concept_name)

    def __getitem__(self, concept_names: Union[str, List[str]]) -> 'SrElement':
        if isinstance(concept_names, str):
            return self.get(concept_names)

        return self.get(*concept_names)

    def __truediv__(self, concept_name) -> 'SrElement':
        return self.get(concept_name)

    def __floordiv__(self, concept_name) -> Iterator['SrElement']:
        return self.get_all(concept_name)

    # Jupyter integration.
    def _repr_html_(self) -> str:
        header = ('Code Value', 'Coding Scheme Designator', 'Code Meaning', 'Value')
        table_rows = [('-', '-', '-', self.value if self.value is not None else '')]
        
        for entry in self.content_it():
            concept_name = entry.concept_name
            value = '' if entry.value is None else entry.value
            table_rows.append((concept_name.code_value,
                               concept_name.coding_scheme_designator,
                               concept_name.code_meaning,
                               value))

        return reper_html(header, table_rows)

    def _ipython_key_completions_(self) -> List[str]:
        keys = []
        for c in self.content_it():
            cn = c.concept_name
            keys.extend([cn.code_meaning, cn.code_value_csd])

        return keys

    # Properties.
    @property
    def concept_name(self) -> ConceptName:
        '''
            Concept name of this instance.
        '''
        assert_concept_name_code_sequence(self._dicom_data)
        concept_name_code = self._dicom_data.ConceptNameCodeSequence[0]

        return ConceptName(code_value=concept_name_code.CodeValue,
                           code_meaning=concept_name_code.CodeMeaning,
                           coding_scheme_designator=concept_name_code.CodingSchemeDesignator)

    @property
    def content(self) -> List['SrElement']:
        '''
            Lists of all entries.
        '''
        return list(self.content_it())

    @property
    def value(self):
        '''
            Extracts the value if a value of supported data types exists. Otherwise, None is returned.
        '''
        return self.get_value()

    # Methods.
    def get_value(self):
        for ve in _VALUE_EXTRACTORS:
            if ve.can_extract(self._dicom_data):
                return ve.extract_value(self._dicom_data)

        return None

    def get(self, *concept_names: str) -> 'SrElement':
        '''
            Returns the first entry with the specified concept name.
            If no entry is found, a ValueError exception is raised.
        '''
        if len(concept_names) > 1:
            return self._get_first(concept_names[0]).get(*concept_names[1:])

        if len(concept_names) == 1:
            return self._get_first(concept_names[0])

        raise ValueError('Invalid argument.')

    def _get_first(self, concept_name: str) -> 'SrElement':
        for concept in self.get_all(concept_name):
            return concept

        message = 'Could not find: "{0}"'.format(concept_name)
        raise SrMissingContentException(message=message)

    def get_all(self, concept_name: str) -> Iterator['SrElement']:
        '''
            Returns an iterator that iterates over entries with the specified concept name.
        '''
        for sr_element in self.content_it():
            cn = sr_element.concept_name
            if concept_name in (cn.code_meaning, cn.code_value_csd):
                yield sr_element

    def content_it(self) -> Iterator['SrElement']:
        '''
            Iterates over all entries.
        '''
        if hasattr(self._dicom_data, 'ContentSequence'):
            return (SrElement(content) for content in self._dicom_data.ContentSequence)

        return iter([])

def assert_concept_name_code_sequence(dicom_data) -> None:
    assert hasattr(dicom_data, 'ConceptNameCodeSequence')
    assert len(dicom_data.ConceptNameCodeSequence) == 1



from rdsr_navigator.value_extractors.value_extractor_base import ValueExtractorBase

class MeasuredValuesExtractor(ValueExtractorBase):

    @staticmethod
    def can_extract(dicom_obj):
        if not hasattr(dicom_obj, 'MeasuredValueSequence'):
            return False
        
        if len(dicom_obj.MeasuredValueSequence) != 1:
            return False
        
        mv = dicom_obj.MeasuredValueSequence[0]
        if not hasattr(mv, 'MeasurementUnitsCodeSequence'):
            return False

        if not hasattr(mv, 'NumericValue'):
            return False

        mucs = mv.MeasurementUnitsCodeSequence
        if len(mucs) != 1:
            return False

        if not hasattr(mucs[0], 'CodeMeaning'):
            return False

        return True

    @staticmethod
    def extract_value(dicom_obj):
        mv = dicom_obj.MeasuredValueSequence[0]
        return (convert_to_number(mv.NumericValue), mv.MeasurementUnitsCodeSequence[0].CodeMeaning)

def convert_to_number(num):
    return float(num)



from typing import Union
from pathlib import Path
import pydicom
from pydicom.dataset import FileDataset
from rdsr_navigator.sr_element import SrElement

def read_file(file: Union[str, Path, FileDataset]):
    '''
        Reads the RDSR content of a DICOM file.

        Parameters
        ----------
        file:
            Path to an RDSR file or pydicom FileDataset.
    '''
    
    if isinstance(file, str):
        return SrElement(pydicom.read_file(file))
    
    if isinstance(file, Path):
        return SrElement(pydicom.read_file(str(file)))

    if isinstance(file, FileDataset):
        return SrElement(file)

    raise TypeError('Unsupported type. Supported types: str, pathlib.Path, or pydicom.dataset.FileDataset')



from rdsr_navigator.value_extractors.value_extractor_base import ValueExtractorBase

class CodeMeaningExtractor(ValueExtractorBase):

    @staticmethod
    def can_extract(dicom_obj):

        if not hasattr(dicom_obj, 'ConceptCodeSequence'):
            return False

        if len(dicom_obj.ConceptCodeSequence) != 1:
            return False

        if not hasattr(dicom_obj.ConceptCodeSequence[0], 'CodeMeaning'):
            return False

        return True

    @staticmethod
    def extract_value(dicom_obj):
        code_sequence = dicom_obj.ConceptCodeSequence
        return code_sequence[0].CodeMeaning



from typing import Sequence

def reper_html(headers: Sequence[str], rows: Sequence[Sequence[object]]) -> str:
    html = [r'<table><thead><tr>', ]

    for header in headers:
        html.append(r'<th>{0}</th>'.format(header))
    html.append(r'</tr></thead><tbody>')

    for row in rows:
        html.append(r'<tr>')
        for value in row:
            html.append(r'<td>{0}</td>'.format(value))
        html.append(r'</tr>')

    html.append(r'</tbody></table>')
    return ''.join(html)



from rdsr_navigator.value_extractors.datetime_extractor import DateTimeExtractor
from rdsr_navigator.value_extractors.measured_values_extractor import MeasuredValuesExtractor
from rdsr_navigator.value_extractors.text_value_extractor import TextValueExtractor
from rdsr_navigator.value_extractors.uid_extractor import UidExtractor
from rdsr_navigator.value_extractors.value_extractor_base import ValueExtractorBase
from rdsr_navigator.value_extractors.code_meaning_extractor import CodeMeaningExtractor



from rdsr_navigator.value_extractors.value_extractor_base import ValueExtractorBase

class DateTimeExtractor(ValueExtractorBase):

    @staticmethod
    def can_extract(dicom_obj):
        return hasattr(dicom_obj, 'DateTime')

    @staticmethod
    def extract_value(dicom_obj):
        return dicom_obj.DateTime



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='rdsr_navigator',
    version='0.2.7',
    author='Robert Vorbau',
    author_email='robert.vorbau@sll.se',
    description='Package for extracting data from DICOM RDSR files',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=r'https://bitbucket.org/robertvorbau_ks/rdsr_navigator/',
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=['pydicom', ],
    classifiers=[
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Physics',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6'
)
